{-# LANGUAGE OverloadedStrings #-}

module BNFC.Backend.Haskell.Parser where

import BNFC.Prelude

import Control.Monad.State

import Prettyprinter

import           Data.List   (intersperse)
import qualified Data.Map    as Map
import           Data.String (fromString)

import System.FilePath (takeBaseName)

import BNFC.Backend.Common.StringUtils (escapeChars)
import BNFC.Backend.Common.Utils as Utils

import BNFC.Backend.CommonInterface.Backend

import BNFC.Backend.Haskell.Options
import BNFC.Backend.Haskell.State
import BNFC.Backend.Haskell.Utilities.Parser
import BNFC.Backend.Haskell.Utilities.Utils

import BNFC.CF
import BNFC.Options.GlobalOptions

haskellParser :: LBNF -> State HaskellBackendState Result
haskellParser lbnf = do
  st <- get
  let cfName      = takeBaseName $ optInput $ globalOpt st
      toks        = lexerParserTokens st
      rules       = parserRules st
      inDirectory = inDir $ haskellOpts st
      nSpace      = nameSpace $ haskellOpts st
      funct       = functor $ haskellOpts st
      tt          = tokenText $ haskellOpts st
      parserSpecification =
        cf2parser lbnf rules cfName inDirectory nSpace tt toks funct
  return [(mkFilePath inDirectory nSpace cfName "Par" "y", parserSpecification)]

cf2parser :: LBNF
          -> [(Cat, Map RHS RuleLabel)]
          -> String
          -> Bool
          -> Maybe String
          -> TokenText
          -> [Token]
          -> Bool
          -> String
cf2parser lbnf rules cfName inDir nameSpace tokenText tokens functor =
  docToString defaultLayoutOptions $ cf2doc lbnf rules cfName inDir nameSpace tokenText tokens functor

cf2doc :: LBNF
       -> [(Cat, Map RHS RuleLabel)]
       -> String
       -> Bool
       -> Maybe String
       -> TokenText
       -> [Token]
       -> Bool
       -> Doc ()
cf2doc lbnf rules cfName inDir nameSpace tokenText tokens functor = (vsep . intersperse emptyDoc)
  [ header parName absName lexName tokenText parsers
  , declarations functor parsers
  -- TODO align tokens.
  , tokensList lbnf tokens functor
  , delimiter
  , specialRules lbnf absName tokenText functor tokens
  , happyRules absName functor rules
  , footer absName usedBuiltins tokenText functor parsers
  ]
  where
    parName :: ModuleName
    parName = mkModule inDir nameSpace cfName "Par"
    absName :: ModuleName
    absName = mkModule inDir nameSpace cfName "Abs"
    lexName :: ModuleName
    lexName = mkModule inDir nameSpace cfName "Lex"
    parsers :: [Cat]
    parsers = fst <$> rules
    delimiter :: Doc ()
    delimiter = "%%"
    usedBuiltins :: [String]
    usedBuiltins = toList . printBuiltinCat <$> Map.keys (_lbnfParserBuiltins lbnf)

header :: ModuleName -> ModuleName -> ModuleName -> TokenText -> [Cat] -> Doc ()
header parName absName lexName tokenText entryPoints = vsep
  [ "-- File generated by the BNF Converter."
  , "-- Parser definition for use with Happy."
  , lbrace
  , "{-# OPTIONS_GHC -fno-warn-incomplete-patterns -fno-warn-overlapping-patterns #-}"
  , "{-# LANGUAGE PatternSynonyms #-}"
  , emptyDoc
  , "module" <+> fromString parName
  , indent 2 "( happyError"
  , indent 2 ", myLexer"
  , vsep $ map (indent 2 . (<>) ", " . parserCatName) entryPoints
  , indent 2 ") where"
  , emptyDoc
  , "import Prelude"
  , emptyDoc
  , "import qualified" <+> fromString absName
  , "import" <+> fromString lexName
  , tokenTextImport tokenText
  , rbrace
  ]

-- | The declarations of a happy file.
-- >>> declarations False [Cat "A", Cat "B", ListCat (Cat "B")]
-- %name pA A
-- %name pB B
-- %name pListB ListB
-- -- no lexer declaration
-- %monad { Err } { (>>=) } { return }
-- %tokentype {Token}
--
-- >>> declarations True [Cat "A", Cat "B", ListCat (Cat "B")]
-- %name pA_internal A
-- %name pB_internal B
-- %name pListB_internal ListB
-- -- no lexer declaration
-- %monad { Err } { (>>=) } { return }
-- %tokentype {Token}
declarations :: Bool -> [Cat] -> Doc ()
declarations functor cats = vsep
  [ vsep $ map (generateP functor) cats
  , "%monad { Err } { (>>=) } { return }"
  , "%tokentype" <+> braces tokenName
  ]

-- | Generate the list of tokens and their identifiers.
tokensList :: LBNF -> [Token] -> Bool -> Doc ()
tokensList lbnf tokens functor
  -- Andreas, 2019-01-02: "%token" followed by nothing is a Happy parse error.
  -- Thus, if we have no tokens, do not output anything.
  | Map.null (_lbnfSymbolsKeywords lbnf) && null tokens = emptyDoc
  | otherwise = hang 2 $ vsep
    [ "%token"
    , vsep $ printToken <$> Map.toList (_lbnfSymbolsKeywords lbnf)
    , vsep (printSpecialTokens <$> tokens)
    ]
  where

    -- Print tokens coming from symbols and keywords.
    printToken :: (String1, Int) -> Doc ()
    printToken (token, n) =
      squotes (fromString $ escapeChars $ toList token) <+>
      braces (space <> "PT _" <+> parens ("TS _" <+> fromString (show n)) <> space)

    -- Print tokens coming from builtin categories and user defined tokens.
    printSpecialTokens :: Token -> Doc ()
    printSpecialTokens (Builtin b) = case b of
      BChar    -> "L_charac" <+> braces (" PT _ (TC" <+> if functor then "_)" else "$$" <> ") ")
      BDouble  -> "L_doubl"  <+> braces (" PT _ (TD" <+> if functor then "_)" else "$$" <> ") ")
      BInteger -> "L_integ"  <+> braces (" PT _ (TI" <+> if functor then "_)" else "$$" <> ") ")
      BString  -> "L_quoted" <+> braces (" PT _ (TL" <+> if functor then "_)" else "$$" <> ") ")
    printSpecialTokens Identifier =
      "L_Ident" <+> braces (" PT _ (TV" <+> if functor then "_)" else "$$" <> ") ")
    printSpecialTokens (UserDefined s) =
      "L_" <> tName <+> braces (" PT _ (T_" <> tName <+> posn s <> ") ")
      where
        tName :: Doc ()
        tName = fromString $ toList s
        posn :: CatName -> Doc ()
        posn tk = case Map.lookup tk (_lbnfTokenDefs lbnf) of
          Nothing   -> panic $ "Token " ++ toList s ++ " is not found in _lbnfTokenDefs"
          Just pDef ->
            if isPositionToken pDef || functor
            then "_"
            else "$$"

specialRules :: LBNF -> ModuleName -> TokenText -> Bool -> [Token] -> Doc ()
specialRules lbnf absName tokenText functor tks = (vsep . intersperse emptyDoc) $
  specialRule lbnf absName tokenText functor <$> tks

-- Happy rules for builtin categories, indentifier and user defined tokens.
specialRule :: LBNF -> ModuleName -> TokenText -> Bool -> Token -> Doc ()
specialRule lbnf absName tokenText functor t = case t of
  Builtin b -> case b of
    BChar -> vsep
      [ "Char :: {" <+> mkType tokenCat <+> "}"
      , "Char  : L_charac {" <+> mkBody <+> "}"
      ]
    BDouble -> vsep
      [ "Double :: {" <+> mkType tokenCat <+> "}"
      , "Double  : L_doubl  {" <+> mkBody <+> "}"]
    BInteger -> vsep
      [ "Integer :: {" <+> mkType tokenCat <+> "}"
      , "Integer  : L_integ  {" <+> mkBody <+> "}"
      ]
    BString -> vsep
      [ "String :: {" <+> mkType tokenCat <+> "}"
      , "String  : L_quoted {" <+> mkBody <+> "}"
      ]
  Identifier -> vsep
      [ "Ident :: {" <+> mkType (qualify absName tokenCat) <+> "}"
      , "Ident  : L_Ident {" <+> mkBody <+> "}"
      ]
  UserDefined s -> vsep
      [ fromString (toList s) <+> ":: {" <+> mkType (qualify absName tokenCat) <+> "}"
      , fromString (toList s) <+> ":" <+> fromString (toList ("L_" <> s)) <+> "{" <+> mkBody <+> "}"
      ]
  where
    tokenCat = printTokenName t

    mkType :: String -> Doc ()
    mkType token =
      if functor
      then parens (fromString (qualify absName posType) <> comma <+> fromString token)
      else fromString token

    mkBody :: Doc ()
    mkBody
      | functor   =
          parens ("uncurry" <+> fromString (qualify absName posConstr) <+> "(tokenLineCol $1)"
          <> comma <+> mkVal t)
      | otherwise = mkVal t

    mkVal :: Token -> Doc ()
    mkVal tk = case tk of
      Builtin b -> case b of
        BChar ->
          if functor
          then "(read" <+> stringUnpack "(tokenText $1)" <+> ") :: Char"
          -- Char never has pos.
          else "(read" <+> stringUnpack "$1" <+> ") :: Char"
        BDouble ->
          if functor
          then "(read" <+> stringUnpack "(tokenText $1)" <+> ") :: Double"
          -- Double never has pos.
          else "(read" <+> stringUnpack "$1" <+> ") :: Double"
        BInteger ->
          if functor
          then "(read" <+> stringUnpack "(tokenText $1)" <+> ") :: Integer"
          -- Integer never has pos.
          else "(read" <+> stringUnpack "$1" <+> ") :: Integer"
        BString ->
          if functor
          then stringUnpack "((\\(PT _ (TL s)) -> s) $1)"
          -- String never has pos.
          else stringUnpack "$1"
      Identifier ->
        if functor
        then fromString (qualify absName "Ident") <+> "(tokenText $1)"
        else fromString (qualify absName "Ident") <+> "$1"
      UserDefined s ->
        if functor
        then
          case Map.lookup s (_lbnfTokenDefs lbnf) of
            Nothing   -> panic $ "Token " ++ toList s ++ " is not found in _lbnfTokenDefs"
            Just pDef ->
              if isPositionToken pDef
              then fromString (qualify absName (toList s)) <+> "(mkPosToken $1)"
              else fromString (qualify absName (toList s)) <+> "(tokenText $1)"
        else
          case Map.lookup s (_lbnfTokenDefs lbnf) of
            Nothing   -> panic $ "Token " ++ toList s ++ " is not found in _lbnfTokenDefs"
            Just pDef ->
              if isPositionToken pDef
              then fromString (qualify absName (toList s)) <+> "(mkPosToken $1)"
              else fromString (qualify absName (toList s)) <+> "$1"

    stringUnpack :: String -> Doc ()
    stringUnpack = tokenTextUnpack tokenText

-- Happy rules.
happyRules :: ModuleName
           -> Bool
           -> [(Cat, Map RHS RuleLabel)]
           -> Doc ()
happyRules absName functor =
  vsep .
  intersperse emptyDoc .
  fmap (uncurry (printRule absName functor))

printRule :: ModuleName
          -> Bool
          -> Cat
          -> Map RHS RuleLabel
          -> Doc ()
printRule absName functor category rhs = vsep
  [ catName <+> "::" <+> ruleType
  , hang 2 $ vsep [ catName, vsep $ zipWith (<+>) (":" : repeat "|") ruleCases]
  ]
  where

    catName :: Doc ()
    catName = fromString $ printCatNamePrec' category

    ruleType :: Doc ()
    ruleType =
      if functor
      then lbrace <+> parens (fromString (qualify absName posType) <> comma <+>
        qualifiedCat) <+> rbrace
      else lbrace <+> qualifiedCat <+> rbrace
      where
        qualifiedCat :: Doc ()
        qualifiedCat =
          if isCatList category
          then
            if isCatBuiltin category
            then brackets $ fromString $ printCatName category
            else brackets $ fromString $ qualify absName $ printCatName category
          else fromString $ qualify absName $ printCatName category

    ruleCases :: [Doc ()]
    ruleCases = constructRule functor absName <$> Map.toList rhs


constructRule :: Bool -> ModuleName -> (RHS, RuleLabel) -> Doc ()
constructRule functor absName (rhs, label) =
  pat' <+> lbrace <+> action  <+> rbrace
  where
    rName           = printRuleName rLabel
    rLabel          = ruleLabel label
    (pat, metavars) = generatePatterns functor rhs
    pat'            = fromString pat
    metavars'       = fromString $ unwords metavars

    action :: Doc ()
    action
      | functor   = parens $ actionPos <> comma <+> actionValue
      | otherwise = actionValue

    actionPos :: Doc ()
    actionPos = case rhs of
      []                -> fromString $ absName ++ "." ++ noPosConstr
      (NTerminal _) : _ -> "fst $1"
      (Terminal _) : _  -> "uncurry" <+> qualified posConstr <+> "(tokenLineCol $1)"

    actionValue :: Doc ()
    actionValue
      | isCoercion rLabel = metavars'
      | isList rLabel     =
        if null metavars then qualified rName else qualified rName <+> metavars'
      | functor           =
        if null metavars
        then qualified rName <+> parens actionPos
        else qualified rName <+> parens actionPos <+> metavars'
      | otherwise         =
        if null metavars then qualified rName else qualified rName <+> metavars'

    qualified :: String -> Doc ()
    qualified s
      | isList rLabel = fromString s
      | otherwise     = fromString $ absName ++ "." ++ s


footer :: ModuleName -> [String] -> TokenText -> Bool -> [Cat] -> Doc ()
footer absName usedBuiltins tokenText functor entryPoints = vsep
  [ lbrace
  , emptyDoc
  , "type Err = Either String"
  , emptyDoc
  , "happyError :: [" <> tokenName <> "] -> Err a"
  , "happyError ts = Left $"
  , indent 2 "\"syntax error at \" ++ tokenPos ts ++ "
  , indent 2 "case ts of"
  , indent 4 "[]      -> []"
  , indent 4 "[Err _] -> \" due to lexer error\""
  , indent 4 "t:_     -> \" before `\"" <+> "++" <+> "(prToken t)" <+> "++" <+> "\"'\""
  , emptyDoc
  , "myLexer ::" <+> tokenTextType tokenText <+> "-> [" <> tokenName <> "]"
  , "myLexer = tokens"
  , vsep $
    emptyDoc
    :
    Utils.when functor
      [ "-- Entrypoints"
      , emptyDoc
      , (vsep . intersperse emptyDoc) $ mkParseFun <$> entryPoints
      ]
  , rbrace
  ]
  where
    mkParseFun :: Cat -> Doc ()
    mkParseFun c = vsep
      [ parserCatName c <+> "::" <+> brackets tokenName <+> "-> Err" <+> c'
      , parserCatName c <+> "=" <+> "fmap snd" <+> "." <+> parserCatName c <> "_internal"
      ]
      where
        c' :: Doc ()
        c' =
          if isCatList c
          then
            if catName `elem` usedBuiltins
            then brackets $ fromString catName
            else brackets $ fromString (qualify absName catName)
          else fromString (qualify absName (printTypeName (catToType c)))
        catName :: String
        catName = printTypeName (catToType c)
